@@ -9,7 +9,7 @@
 /* of which you should have received a copy together with this file.        */
 /****************************************************************************/
 
-#define VERSION "2.3"
+#define VERSION "4.0"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,8 +30,9 @@
 #include <string.h>
 
 // Set VARIANTS for in WinBoard variant feature. (With -s option this will always be reset to use "shogi".)
-#define VARIANTS "normal,xiangqi"
-#define STDVARS "chess,chess960,crazyhouse,threecheck,giveaway,atomic,seirawan,shogi,xiangqi"
+#define VARIANTS ",normal,xiangqi"
+#define STDVARS "chess,chess960,crazyhouse,3check,giveaway,suicide,losers,atomic,seirawan,shogi,xiangqi"
+#define EGT ",gaviotaTbPath,syzygyPath,nalimovPath,robbotripleBaseDirectory,robbototalBaseDirectory,bitbases path,"
 
 #define DPRINT if(debug) printf
 #define EPRINT(X) { char f[999]; sprintf X; DPRINT("%s", f); fprintf(toE, "%s", f + 2*(*f == '#')); /* strip optional # prefix */ }
@@ -41,20 +42,20 @@
 #define NONE  2
 #define ANALYZE 3
 
-char move[2000][10], checkOptions[8192], iniPos[256], hashOpt[20], pause, pondering, suspended, ponder, post, hasHash, c, sc='c', suffix[81], *variants, varOpt;
+char move[2000][10], iniPos[256], hashOpt[20], suspended, ponder, post, hasHash, c, sc='c', suffix[81], varOpt, searching, *binary;
 int mps, tc, inc, sTime, depth, myTime, hisTime, stm, computer = NONE, memory, oldMem=0, cores, moveNr, lastDepth, lastScore, startTime, debug, flob;
 int statDepth, statScore, statNodes, statTime, currNr, size, collect, nr, sm, inex, on[500], frc, byo = -1, namOpt, comp;
-char currMove[20], moveMap[500][10], /* for analyze mode */ canPonder[20], threadOpt[20], varList[8000];
-char board[100];  // XQ board for UCCI
+char currMove[20], moveMap[500][10], /* for analyze mode */ canPonder[20], threadOpt[20], varList[8000], anaOpt[20], checkOptions[8192] = "Ponder";
+char pvs[99][999], board[100];  // XQ board for UCCI
 char *nameWord = "name ", *valueWord = "value ", *wTime = "w", *bTime = "b", *wInc = "winc", *bInc = "binc", newGame; // keywords that differ in UCCI
-int unit = 1, drawOffer;
+int unit = 1, drawOffer, scores[99], mpvSP, maxDepth, ponderAlways;
 
 FILE *toE, *fromE, *fromF;
 int pid;
 
-#ifdef WIN32
-char *strcasestr (char *p, char *q) { char *r = p; while(*r) *r = tolower(*r), r++; return strstr(p, q); }
+char *strcasestr (char *p, char *q) { while(*p) { char *r=p++, *s=q; while(tolower(*r++) == tolower(*s) && *s) s++; if(!*s) return p-1; } return NULL; }
 
+#ifdef WIN32
 WinPipe(HANDLE *hRd, HANDLE *hWr)
 {
   SECURITY_ATTRIBUTES saAttr;
@@ -155,16 +156,15 @@ StartSearch(char *ponder)
 	if(inc && !*suffix) EPRINT((f, " %s %d %s %d", wInc, inc, bInc, inc))
 	if(depth > 0) EPRINT((f, " depth %d", depth))
         if(*suffix) EPRINT((f, suffix, inc))
-	EPRINT((f, "\n"))
+	EPRINT((f, "\n")); maxDepth = mpvSP = 0;
 }
 
 void
-StopPonder(int pondering)
+StopSearch(int discard)
 {
-	if(!pondering) return;
-	pause = 1;
+	if(!searching) return;
+	if(discard) searching = 0; // this causes bestmove to be ignored
 	EPRINT((f, "# stop\n")) fflush(toE); // note: 'pondering' remains set until engine acknowledges 'stop' with 'bestmove'
-	Sync(PAUSE); // wait for engine to acknowledge 'stop' with 'bestmove'.
 }
 
 void
@@ -183,14 +183,21 @@ LoadPos(int moveNr)
 }
 
 void
-StartPonder()
+StartPonder(int moveNr)
 {
 	if(!move[moveNr][0]) return; // no ponder move
 	LoadPos(moveNr+1);
-	pondering = 1; lastDepth = 1;
+	searching = 1; lastDepth = 1;
+	DPRINT("# ponder on %s\n", move[moveNr]);
 	StartSearch(" ponder");
 }
 
+void
+Analyze(char *val)
+{
+    if(*anaOpt) EPRINT((f, "# setoption %s%s %s%s\n", nameWord, anaOpt, valueWord, val));
+}
+
 char *Convert(char *pv)
 {   // convert Shogi coordinates to WB
     char *p, *q, c;
@@ -224,34 +231,29 @@ Move4GUI(char *m)
 }
 
 int
-GetChar()
+ReadLine (FILE *f, char *line)
 {
-    int c;
-    if(fromF) {
-	if((c = fgetc(fromF)) != EOF) return c;
-	fclose(fromF); fromF = 0; printf("# end fake\n");
-    }
-    return fgetc(fromE);
+    int x, i = 0;
+    while((x = fgetc(f)) != EOF && (line[i] = x) != '\n') i++; line[++i] = 0;
+    return (x != EOF);
 }
 
-void *
-Engine2GUI()
+void
+HandleEngineOutput()
 {
-    char line[1024], command[256];
+    char line[1024], command[256]; static char egts[999];
 
-    if(fromF = fopen("DefectiveEngineOptions.ini", "r")) printf("# fake engine input\n");
     while(1) {
-	int i=0, x; char *p, dummy;
+	int i=0, x; char *p, dummy, len;
 
 	fflush(stdout); fflush(toE);
-	while((line[i] = x = GetChar()) != EOF && line[i] != '\n') i++;
-	line[++i] = 0;
-	if(x == EOF) exit(0);
+	if(fromF && !ReadLine(fromF, line))  fromF = 0, printf("# end fake\n");
+	if(!fromF && !ReadLine(fromE, line)) printf("tellusererror UCI2WB: %s died on me\n", binary), exit(0);
 	DPRINT("# engine said: %s", line), fflush(stdout);
 	if(sscanf(line, "%s", command) != 1) continue;
 	if(!strcmp(command, "bestmove")) {
-	    if(pause == 1) { pondering = pause = 0; Sync(WAKEUP); continue; } // bestmove was reply to ponder miss or analysis result; ignore.
-	    else if(pondering) { pondering = 0; printf("%d 0 0 0 UCI violation! Engine moves during ponder\n", lastDepth+1); continue; } // ignore ponder search
+	    if(searching == 1) { searching = 0; printf("%d 0 0 0 UCI violation! Engine moves during ponder\n", lastDepth+1); return; } // ignore ponder search
+	    else if(searching != 3) { searching = 0; return; } // ponder miss or analysis result; ignore.
 	    // move was a move to be played
 	    if(p = strstr(line+8, " draw")) *p = 0, printf("offer draw\n"); // UCCI
 	    if(strstr(line+9, "resign")) { printf("resign\n"); computer = NONE; }
@@ -259,42 +261,36 @@ Engine2GUI()
 	    if(strstr(line+9, "(none)") || strstr(line+9, "null") ||
 	       strstr(line+9, "0000")) { printf("%s\n", lastScore < -99999 ? "resign" : "1/2-1/2 {stalemate}"); computer = NONE; }
 	    sscanf(line, "bestmove %s", move[moveNr++]);
-	    myTime -= (GetTickCount() - startTime)*1.02 + inc; // update own clock, so we can give correct wtime, btime with ponder
+	    myTime -= (GetTickCount() - startTime)*1.02 - inc; // update own clock, so we can give correct wtime, btime with ponder
 	    if(mps && ((moveNr+1)/2) % mps == 0) myTime += tc; if(sTime) myTime = sTime; // new session or move starts
-	    stm = WHITE+BLACK - stm;
+	    stm = WHITE+BLACK - stm; searching = 0;
 	    // first start a new ponder search, if pondering is on and we have a move to ponder on
 	    if(p = strstr(line+9, "ponder")) {
 	      sscanf(p+7, "%s", move[moveNr]);
-	      if(computer != NONE && ponder) {
-		DPRINT("# ponder on %s\n", move[moveNr]);
-		StartPonder();
-	      }
+	      if(computer != NONE && ponder) StartPonder(moveNr);
 	      p[-1] = '\n'; *p = 0; // strip off ponder move
 	    } else move[moveNr][0] = 0;
 	    Move4GUI(line+9);
 	    printf("move %s\n", line+9); // send move to GUI
             if(move[moveNr][0]) printf("Hint: %s\n", move[moveNr]);
-	    if(pause) { pause = 0; Sync(WAKEUP); } // release commands that came in during think
 	    if(lastScore == 100001 && iniPos[0] != 'f') { printf("%s {mate}\n", stm == BLACK ? "1-0" : "0-1"); computer = NONE; }
+	    fflush(stdout); return;
 	}
 	else if(!strcmp(command, "info")) {
 	    int d=0, s=0, t=(GetTickCount() - startTime)/10, n=1;
 	    char *pv, varName[80];
 	    if(sscanf(line+5, "string times @ %c", &dummy) == 1) { printf("# %s", line+12); continue; }
 	    if(sscanf(line+5, "string variant %s", varName) == 1) {
-		if(!strstr(STDVARS, varName) && (p = strstr(line+18, " startpos "))) printf("setup (-) 8x8+0_fairy %s", p+10);
-		continue;
-	    }
-	    if(collect && (pv = strstr(line+5, "currmove "))) {
-		if(p = strstr(line+5, "currmovenumber ")) {
-		    n = atoi(p+15);
-		    if(collect == 1 && n != 1) continue; // wait for move 1
-		    if(collect + (n == 1) > 2) { // done collecting
-			if(inex && collect == 2) printf("%d 0 0 0 OK to exclude\n", lastDepth);
-			collect = 3; continue;
-		    }
-		    collect = 2; on[nr=n] = 1; sscanf(pv+9, "%s", moveMap[n]); continue; // store move
+		if(!strstr(STDVARS, varName)) {
+		    int files = 8, ranks = 8, hand = 0; char parent[80];
+		    if(p = strstr(line+18, " files ")) sscanf(p+7, "%d", &files);
+		    if(p = strstr(line+18, " ranks ")) sscanf(p+7, "%d", &ranks);
+		    if(p = strstr(line+18, " pocket ")) sscanf(p+8, "%d", &hand);
+		    if(p = strstr(line+18, " template ")) sscanf(p+10, "%s", parent); else strcpy(parent, "fairy");
+		    if(p = strstr(line+18, " startpos "))
+			printf("setup (-) %dx%d+%d_%s %s", files, ranks, hand, parent, p+10);
 		}
+		continue;
 	    }
 	    if(!post) continue;
 	    if(sscanf(line+5, "string %c", &dummy) == 1) printf("%d 0 0 0 %s", lastDepth, line+12); else {
@@ -304,15 +300,31 @@ Engine2GUI()
 		if(p = strstr(line+4, " score "))      sscanf(p+7, "%d", &s), statScore = s;
 		if(p = strstr(line+4, " nodes "))      sscanf(p+7, "%d", &n), statNodes = n;
 		if(p = strstr(line+4, " time "))       sscanf(p+6, "%d", &t), t /= 10, statTime = t;
-		if(p = strstr(line+4, " currmove "))   sscanf(p+10,"%s", currMove);
-		if(p = strstr(line+4, " currmovenumber ")) sscanf(p+16,"%d", &currNr);
-		if(pv = strstr(line+4, " pv ")) // convert PV info to WB thinking output
+		if(pv = strstr(line+4, " pv ")) { // convert PV info to WB thinking output
+		  if(d > maxDepth) maxDepth = d, mpvSP = 0; else if(d < maxDepth) continue; // ignore depth regressions
+		  if(p = strstr(line+4, " upperbound ")) strcat(p, "?\n"); else
+		  if(p = strstr(line+4, " lowerbound ")) strcat(p, "!\n");
+		  for(i=0; i<mpvSP; i++) if(s == scores[i] && !strcmp(pvs[i], pv+4)) break; // check if duplicat
+		  if(i >= mpvSP) strncpy(pvs[mpvSP], pv+4, 998), scores[mpvSP++] = s,       // emit as thinking output if not
 		    printf("%3d  %6d %6d %10d %s", lastDepth=d, lastScore=s, t, n, Convert(pv+4));
+		} else if(s == -100000) lastScore = s; // when checkmated score is valid even without PV (which might not come)
+	    }
+	    if(collect && (pv = strstr(line+4, " currmove "))) {
+		sscanf(pv+10,"%s", currMove);
+		if(p = strstr(line+4, " currmovenumber ")) {
+		    n = currNr = atoi(p+16);
+		    if(collect == 1 && n != 1) continue; // wait for move 1
+		    if(collect + (n == 1) > 2) { // done collecting
+			if(inex && collect == 2) printf("%d 0 0 0 OK to exclude\n", lastDepth);
+			collect = 3; continue;
+		    }
+		    collect = 2; on[nr=n] = 1; strcpy(moveMap[n], currMove); continue; // store move
+		}
 	    }
 	}
 	else if(!strcmp(command, "option")) { // USI option: extract data fields
 	    char name[80], type[80], buf[1024], val[256], *q;
-	    int min=0, max=1e9;
+	    int min=0, max=1e9; *val = 0;
 	    if(p = strstr(line+6, " type ")) sscanf(p+1, "type %s", type), *p = '\n';
 	    if(p = strstr(line+6, " min "))  sscanf(p+1, "min %d", &min), *p = '\n';
 	    if(p = strstr(line+6, " max "))  sscanf(p+1, "max %d", &max), *p = '\n';
@@ -321,6 +333,7 @@ Engine2GUI()
 	    if(!strcasecmp(name, "UCI_Chess960")) { frc=2; continue; }
 	    if(!strcasecmp(name, "UCI_Variant")) { if(p = strstr(line+6, " var ")) strcpy(varList, p); varOpt = 1; continue; }
 	    if(!strcasecmp(name, "UCI_Opponent")) { namOpt = 1; continue; }
+	    if(!strcasecmp(name+2, "I_AnalyseMode")) { strcpy(anaOpt, name); continue; }
 	    if(frc< 0 && (strstr(name, "960") || strcasestr(name, "frc")) && !strcmp(type, "check")) {
 		EPRINT((f, "# setoption name %s value true\n", name)) strcpy(val, "true"); // set non-standard suspected FRC options
 	    }
@@ -333,8 +346,15 @@ Engine2GUI()
 	    }
 	    if(!strcasecmp(name, "newgame") && !strcmp(type, "button")) { newGame++; continue; }
 	    if(!strcasecmp(name, "usemillisec")) { unit = (!strcmp(val, "false") ? 2 : 1); continue; }
+	    sprintf(buf, ",%s,", name); if(p = strcasestr(EGT, buf)) { // collect EGT formats
+		strcpy(buf, p); for(p=buf; *++p >='a';){} if(*p == ' ') strcpy(buf, ",scorpio"); *p = 0; strcat(egts, buf); continue; // clip at first non-lower-case
+	    }
 	    // pass on engine-defined option as WB option feature
 	    if(!strcmp(type, "filename")) type[4] = 0;
+	    else if(sc == 'c' && !strcmp(type, "string")) { // in UCI try to guess which strings are file or directory names
+		if(strcasestr(name, "file")) strcpy(type, "file"); else
+		if(strcasestr(name, "path") || strcasestr(name, "directory") || strcasestr(name, "folder")) strcpy(type, "path");
+	    }
 	    sprintf(buf, "feature option=\"%s -%s", name, type); q = buf + strlen(buf);
 	    if(     !strcmp(type, "file")
 	         || !strcmp(type, "string")) sprintf(q, " %s\"\n", val);
@@ -349,6 +369,7 @@ Engine2GUI()
 		    sprintf(buf + strlen(buf), "%s%s%s", min++ ? " /// " : " ", strcmp(type, val) ? "" : "*", val);
 		}
 		strcat(q, "\"\n");
+
 	    }
 	    else buf[0] = 0; // ignore unrecognized option types
 	    if(buf[0]) printf("%s", buf);
@@ -358,18 +379,19 @@ Engine2GUI()
 	    if(sscanf(line, "id name %[^\n]", name) == 1) printf("feature myname=\"%s (U%cI2WB)\"\n", name, sc-32);
 	    if(sscanf(line, "id version %[^\n]", version) == 1 && *name) printf("feature myname=\"%s %s (U%cI2WB)\"\n", name, version, sc-32);
 	}
-	else if(!strcmp(command, "readyok")) { pause = 0; Sync(WAKEUP); } // resume processing of GUI commands
+	else if(!strcmp(command, "readyok")) return; // resume processing of GUI commands
 	else if(sc == 'x'&& !strcmp(command, "ucciok") || sscanf(command, "u%ciok", &c)==1 && c==sc) {
 	    char *p = varList, *q = varList;
 	    while(*q && *q != '\n')  if(!strncmp(q, " var ", 5)) *p++ = ',', q +=5; // replace var keywords by commas
 				else if(!strncmp(q-1, " chess ", 7)) strcpy(p, "normal"), p += 6, q += 5; // 'chess' is called 'normal' in CECP
-				else if(!strncmp(q-1, " threecheck", 11)) *p++ = '3', q += 5; // 'threecheck' is written '3check' in CECP
 				else *p++ = *q++; // copy other variant names unmodified
 	    if(frc) sprintf(p, ",normal,fischerandom"), printf("feature oocastle=%d\n", frc<0); // unannounced FRC uses O-O castling
-	    if(*varList) printf("feature variants=\"%s\"\n", varList+1); // from UCI_Variant combo and/or UCI_Chess960 check options
+	    if(!*varList) strcpy(varList, sc=='s' ? ",shogi,5x5+5_shogi" : VARIANTS); // without clue guess liberally
+	    printf("feature variants=\"%s\"\n", varList+1); // from UCI_Variant combo and/or UCI_Chess960 check options
+	    if(*egts) printf("feature egt=\"%s\"\n", egts+1);
 	    printf("feature smp=1 memory=%d done=1\n", hasHash);
 	    if(unit == 2) { unit = 1; EPRINT((f, "# setoption usemillisec true\n")) }
-	    Sync(WAKEUP); // done with options
+	    fflush(stdout); return; // done with options
 	}
     }
 }
@@ -392,62 +414,152 @@ Move4Engine(char *m)
     }
 }
 
+void DoCommand ();
+char mySide;
+volatile char queue[10000], *qStart, *qEnd;
+
 void
-GUI2Engine()
+LaunchSearch()
 {
-    char line[256], command[256], *p, *q, *r, mySide, searching = 0;
+    int i;
 
-    while(1) {
-	int i, x;
+    if(suspended || searching) return;
 
-	if((computer == stm || computer == ANALYZE && !searching) && !suspended) {
+	if(computer == stm || computer == ANALYZE && sm != 1) {
 	    DPRINT("# start search\n");
 	    LoadPos(moveNr); fflush(stdout); // load position
 	    // and set engine thinking (note USI swaps colors!)
 	    startTime = GetTickCount(); mySide = stm; // remember side we last played for
 	    if(computer == ANALYZE) {
-		EPRINT((f, "# go infinite"))
+		EPRINT((f, "# go infinite")); maxDepth = mpvSP = 0;
 		if(sm & 1) { // some moves are disabled
 		    EPRINT((f, " searchmoves"))
 		    for(i=1; i<nr; i++) if(on[i]) EPRINT((f, " %s", moveMap[i]))
 		}
-		EPRINT((f, "\n")) searching = 1; // suppresses spurious commands during analysis starting new searches
-	    } else { pause = 2; StartSearch(""); fflush(stdout); fflush(toE); Sync(PAUSE); } // block input during thinking
-	}
-      nomove:
+		EPRINT((f, "\n")) searching = 2; // suppresses spurious commands during analysis starting new searches
+	    } else searching = 3, StartSearch(""); // request suspending of input processing while thinking
+	} else if(ponderAlways && computer == NONE) move[moveNr][0] = 0, StartPonder(moveNr-1);
+	else if(BLACK+WHITE-stm == computer && ponder && moveNr) StartPonder(moveNr);
+}
+
+void
+GUI2Engine()
+{
+    char line[256], command[256], *p;
+
+    while(1) {
+	int i, difficult;
+
+       for(difficult=0; !difficult; ) { // read and handle commands that can (or must) be handled during thinking
 	fflush(toE); fflush(stdout);
-	i = 0; while((x = getchar()) != EOF && (line[i] = x) != '\n') i++;
-	line[++i] = 0; if(x == EOF) { printf("# EOF\n"); EPRINT((f, "# quit\n")) exit(-1); }
+	if(!ReadLine(stdin, line)) printf("# EOF\n"), sprintf(line, "quit -1\n");
 	sscanf(line, "%s", command);
+	if(!strcmp(command, "usermove")) { difficult--; break; } // for efficiency during game play, moves, time & otim are tried first
+	else if(!strcmp(command, "time"))   sscanf(line+4, "%d", &myTime),  myTime  = (10*myTime)/unit;
+	else if(!strcmp(command, "otim"))   sscanf(line+4, "%d", &hisTime), hisTime = (10*hisTime)/unit;
+	else if(!strcmp(command, "offer")) drawOffer = 1; // backlogged anyway, so this can be done instantly
+	else if(!strcmp(command, "post"))  post = 1;
+	else if(!strcmp(command, "nopost"))post = 0;
+	else if(!strcmp(command, ".")) {
+	    printf("stat01: %d %d %d %d %d %s\n", statTime, statNodes, statDepth, nr-currNr, nr, currMove);
+	}
+	else if(!strcmp(command, "pause")) {
+	    if(computer == stm) myTime -= GetTickCount() - startTime;
+	    suspended = 1 + (searching == 1); // remember if we were pondering, and stop search ignoring bestmove
+	    StopSearch(1);
+	}
+	else if(!strcmp(command, "xboard")) ;
+	else if(!strcmp(command, "random")) ;
+	else if(!strcmp(command, "accepted")) ;
+	else if(!strcmp(command, "rejected")) ;
+	else if(!strcmp(command, "book")) ;
+	else if(!strcmp(command, "ics")) ;
+	else if(!strcmp(command, "hint")) ;
+	else if(!strcmp(command, "computer")) comp = 1;
+	else { //convert easy & hard to "option" after treating their effect on the adapter
+	  if(!strcmp(command, "easy")) {
+	    if(*canPonder) ponder = 0, sprintf(command, "option"), sprintf(line, "option %s=0\n", canPonder); else continue;
+	  }
+	  else if(!strcmp(command, "hard")) {
+	    if(*canPonder) ponder = 1, sprintf(command, "option"), sprintf(line, "option %s=1\n", canPonder); else continue;
+	  }
+	  if(!strcmp(command, "option")) {
+	    if(sscanf(line+7, "UCI2WB debug output=%d", &debug) == 1) ; else
+	    if(sscanf(line+7, "ponder always=%d", &ponderAlways) == 1) ; else
+	    if(sscanf(line+7, "Floating Byoyomi=%d", &flob) == 1) ; else
+	    if(sscanf(line+7, "Byoyomi=%d", &byo) == 1) ; else
+	    difficult = 1;
+	  }
+	  else difficult = 1; // difficult command; terminate loop for easy ones
+	}
+       } // next command
+
+	// some commands that should never come during thinking can be safely processed here
+	if(difficult < 0) { // used as kludge to signal "usermove" was already matched
+	    sscanf(line, "usermove %s", command); // strips off linefeed
+	    Move4Engine(command);
+	    stm = WHITE+BLACK - stm; collect = (computer == ANALYZE); sm = 0;
+	    // when pondering we either continue the ponder search as normal search, or abort it
+	    if(searching) { // move cannot come during think, so we are pondering or analysing
+		if(searching == 1 && !strcmp(command, move[moveNr])) { // ponder hit
+		    char *draw = drawOffer ? " draw" : ""; drawOffer = 0;
+		    searching = 3; moveNr++; startTime = GetTickCount(); // clock starts running now
+		    EPRINT((f, "# ponderhit%s\n", draw)) fflush(toE); fflush(stdout);
+		    continue;
+		}
+		StopSearch(1);
+	    }
+	    strcpy(move[moveNr++], command); // possibly overwrites ponder move
+	    *qEnd++ = '\n'; Sync(WAKEUP);    // make sure engine thread considers starting a search
+	} else
+	if(!strcmp(command, "resume")) {
+	    if(suspended == 2) StartPonder(moveNr); // restart interrupted ponder search
+	    suspended = 0;  *qEnd++ = '\n'; Sync(WAKEUP); // causes search to start in normal way if on move or analyzing
+	} else
+      {
+	DPRINT("# queue '%s', searching=%d\n", command, searching);
+	if(searching == 3) { // command arrived during thinking; order abort for 'instant commands'
+	    if(!strcmp(command, "?") || !strcmp(command, "quit") ||
+	       !strcmp(command, "force") || !strcmp(command, "result")) StopSearch(0);
+	} else StopSearch(1); // always abort pondering or analysis
+
+	// queue command for execution by engine thread
+	if(qStart == qEnd) qStart = qEnd = queue;
+	p = line; while(qEnd < queue+10000 && (*qEnd++ = *p++) != '\n') {}
+	Sync(WAKEUP);
+      }
+    }
+}
+
+void
+DoCommand ()
+{
+    char line[1024], command[256], *p, *q, *r, type[99];
+    int i;
+
+    p=line; while(qStart < qEnd && (*p++ = *qStart++) != '\n') {} *p = 0;
+    if(line[0] == '\n') return;
+    sscanf(line, "%s", command); DPRINT("# command %s\n", command), fflush(stdout);
+
 	if(!strcmp(command, "new")) {
 	    computer = BLACK; moveNr = 0; depth = -1; move[0][0] = 0;
 	    stm = WHITE; strcpy(iniPos, "position startpos"); frc &= ~1;
 	    if(memory != oldMem && hasHash) EPRINT((f, "# setoption %s%s %s%d\n", nameWord, hashOpt, valueWord, memory))
 	    oldMem = memory;
 	    // we can set other options here
-	    if(sc == 'x') { if(newGame) EPRINT((f, "# setoption newgame\n")) } else // optional in UCCI
 	    if(varOpt) EPRINT((f, "# setoption name UCI_Variant value chess\n"))
-	    pause = 1; // wait for option settings to take effect
 	    EPRINT((f, "# isready\n")) fflush(toE);
-	    Sync(PAUSE); // wait for readyok
+	    HandleEngineOutput(); // wait for readyok
+	    if(sc == 'x') { if(newGame) EPRINT((f, "# setoption newgame\n")) } else // optional in UCCI
 	    EPRINT((f, "# u%cinewgame\n", sc)) fflush(toE);
 	}
-	else if(!strcmp(command, "usermove")) {
-	    sscanf(line, "usermove %s", command); // strips off linefeed
-	    Move4Engine(command);
-	    stm = WHITE+BLACK - stm; collect = (computer == ANALYZE); sm = 0;
-	    // when pondering we either continue the ponder search as normal search, or abort it
-	    if(pondering || computer == ANALYZE) {
-		if(pondering && !strcmp(command, move[moveNr])) { // ponder hit
-		    char *draw = drawOffer ? " draw" : ""; drawOffer = 0;
-		    pondering = 0; pause = 2; moveNr++; startTime = GetTickCount(); // clock starts running now
-		    EPRINT((f, "# ponderhit%s\n", draw)) fflush(toE); fflush(stdout);
-		    Sync(PAUSE); // block input during thinking
-		    goto nomove;
-		}
-		StopPonder(1); searching = 0;
-	    }
-	    strcpy(move[moveNr++], command); // possibly overwrites ponder move
+	else if(!strcmp(command, "option")) {
+	    char *p;
+	    if(p = strchr(line, '=')) {
+		*p++ = 0;
+		if(strstr(checkOptions, line+7)) sprintf(p, "%s\n", atoi(p) ? "true" : "false");
+		EPRINT((f, "# setoption %s%s %s%s", nameWord, line+7, valueWord, p));
+	    } else EPRINT((f, "# setoption %s%s\n", nameWord, line+7));
 	}
 	else if(!strcmp(command, "level")) {
 	    int sec = 0;
@@ -455,25 +567,13 @@ GUI2Engine()
 	    sscanf(line, "level %d %d %d", &mps, &tc, &inc);
 	    tc = (60*tc + sec)*1000; inc *= 1000; sTime = 0; tc /= unit; inc /= unit;
 	}
-	else if(!strcmp(command, "option")) {
-	    char name[80], *p;
-	    if(searching) StopPonder(1), searching = 0; // force new search if settings change during analysis (multi-PV!)
-	    if(sscanf(line+7, "UCI2WB debug output=%d", &debug) == 1) ; else
-	    if(sscanf(line+7, "Floating Byoyomi=%d", &flob) == 1) ; else
-	    if(sscanf(line+7, "Byoyomi=%d", &byo) == 1) ; else
-	    if(p = strchr(line, '=')) {
-		*p++ = 0;
-		if(strstr(checkOptions, line+7)) sprintf(p, "%s\n", atoi(p) ? "true" : "false");
-		EPRINT((f, "# setoption %s%s %s%s", nameWord, line+7, valueWord, p))
-	    } else EPRINT((f, "# setoption %s%s\n", nameWord, line+7))
-	}
 	else if(!strcmp(command, "protover")) {
-	    if(!variants) variants = sc=='s' ? "shogi,5x5+5_shogi" : VARIANTS;
-	    printf("feature variants=\"%s\" setboard=1 usermove=1 debug=1 ping=1 name=1 reuse=0 exclude=1 pause=1 sigint=0 sigterm=0 done=0\n", variants);
+	    printf("feature setboard=1 usermove=1 debug=1 ping=1 name=1 reuse=0 exclude=1 pause=1 sigint=0 sigterm=0 done=0\n");
 	    printf("feature option=\"UCI2WB debug output -check %d\"\n", debug);
+	    printf("feature option=\"ponder always -check %d\"\n", ponderAlways);
 	    if(sc == 's') printf("feature option=\"Floating Byoyomi -check %d\"\nfeature option=\"Byoyomi -spin %d -1 1000\"\n", flob, byo);
-	    EPRINT((f, sc == 'x' ? "# ucci\n" : "u%ci\n", sc)) fflush(toE); // prompt UCI engine for options
-	    Sync(PAUSE); // wait for uciok
+	    EPRINT((f, sc == 'x' ? "# ucci\n" : "# u%ci\n", sc)) fflush(toE); // prompt UCI engine for options
+	    HandleEngineOutput(); // wait for uciok
 	}
 	else if(!strcmp(command, "setboard")) {
 		stm = (strstr(line+9, " b ") ? BLACK : WHITE);
@@ -489,8 +589,8 @@ GUI2Engine()
 	}
 	else if(!strcmp(command, "variant")) {
 		if(varOpt) {
-		    EPRINT((f, "# setoption name UCI_Variant value %sucinewgame\nisready\n", strcmp(line+8, "3check\n") ? line+8 : "threecheck\n"))
-		    fflush(toE); Sync(PAUSE);
+		    EPRINT((f, "# setoption name UCI_Variant value %sucinewgame\nisready\n", line+8))
+		    fflush(toE); HandleEngineOutput(); // wait for readyok
 		}
 		if(!strcmp(line+8, "shogi\n")) size = 9, strcpy(iniPos, "position startpos");
 		if(!strcmp(line+8, "5x5+5_shogi\n")) size = 5, strcpy(iniPos, "position startpos");
@@ -498,50 +598,47 @@ GUI2Engine()
 		if(!strcmp(line+8, "fischerandom\n")) { frc |= 1; if(frc > 0) EPRINT((f, "# setoption name UCI_Chess960 value true\n")) }
 	}
 	else if(!strcmp(command, "undo") && (i=1) || !strcmp(command, "remove") && (i=2)) {
-	    if(pondering || computer == ANALYZE) StopPonder(1), searching = 0;
 	    moveNr = moveNr > i ? moveNr - i : 0; collect = (computer == ANALYZE); sm = 0;
 	}
-	else if(!strcmp(command, ".")) {
-	    printf("stat01: %d %d %d %d 100 %s\n", statTime, statNodes, statDepth, 100-currNr, currMove);
-	    goto nomove;
-	}
 	else if(!strcmp(command+2, "clude") && collect > 2) { // include or exclude
 	    int all = !strcmp(line+8, "all"), in = command[1] == 'n';
 	    inex = 1; line[strlen(line)-1] = sm = 0; // strip LF and clear sm flag
 	    for(i=1; i<nr; i++) { if(!strcmp(line+8, moveMap[i]) || all) on[i] = in; sm |= on[i]+1; } // sm: 2 = enabled, 1 = disabled
-	    if(!(sm & 2)) goto nomove; // no moves enabled; continue current search
-	    if(computer == ANALYZE) StopPonder(1), searching = 0; // abort old analysis
 	}
-	else if(!strcmp(command, "pause")) {
-	    if(computer == stm) myTime -= GetTickCount() - startTime;
-	    suspended = 1 + pondering; // remember if we were pondering, and stop search ignoring bestmove
-	    StopPonder(pondering || computer == stm);
-	}
-	else if(!strcmp(command, "resume")) {
-	    if(suspended == 2) StartPonder(); // restart interrupted ponder search
-	    suspended = 0; // causes thinking to start in normal way if on move or analyzing
-	}
-	else if(!strcmp(command, "xboard")) ;
-	else if(!strcmp(command, "analyze"))computer = ANALYZE, collect = 1, sm = 0;
-	else if(!strcmp(command, "exit"))   computer = NONE, StopPonder(1), searching = 0;
-	else if(!strcmp(command, "force"))  computer = NONE, StopPonder(pondering);
+	else if(!strcmp(command, "analyze"))computer = ANALYZE, collect = 1, sm = 0, Analyze("true");
+	else if(!strcmp(command, "exit"))   computer = NONE, Analyze("false");
+	else if(!strcmp(command, "force"))  computer = NONE;
 	else if(!strcmp(command, "go"))     computer = stm;
-	else if(!strcmp(command, "time"))   sscanf(line+4, "%d", &myTime),  myTime  = (10*myTime)/unit;
-	else if(!strcmp(command, "otim"))   sscanf(line+4, "%d", &hisTime), hisTime = (10*hisTime)/unit;
-	else if(!strcmp(command, "post"))   post = 1;
-	else if(!strcmp(command, "nopost")) post = 0;
-	else if(!strcmp(command, "easy") && !!*canPonder) { ponder = 0; StopPonder(pondering); EPRINT((f, "# setoption %s%s %sfalse\n", nameWord, canPonder, valueWord)) }
-	else if(!strcmp(command, "hard") && !!*canPonder) { ponder = 1; EPRINT((f, "# setoption %s%s %strue\n", nameWord, canPonder, valueWord)) StartPonder(); }
 	else if(!strcmp(command, "ping"))   { /* static int done; if(!done) pause = 1, fprintf(toE, "isready\n"), fflush(toE), printf("# send isready\n"), fflush(stdout), Sync(PAUSE); done = 1;*/ printf("po%s", line+2); }
 	else if(!strcmp(command, "memory")) sscanf(line, "memory %d", &memory);
 	else if(!strcmp(command, "cores")&& !!*threadOpt) { sscanf(line, "cores %d", &cores); EPRINT((f, "# setoption %s%s %s%d\n", nameWord, threadOpt, valueWord, cores)) }
+	else if(!strcmp(command, "egtpath")){
+	    sscanf(line+8, "%s %[^\n]", type, command);
+            if(p = strstr(EGT, type)) strcpy(type, p), p = strchr(type, ','), *p = 0; else strcpy(type, "bitbases path");
+	    EPRINT((f, "# setoption name %s value %s\n", type, command));
+	}
 	else if(!strcmp(command, "sd"))     sscanf(line, "sd %d", &depth);
 	else if(!strcmp(command, "st"))     sscanf(line, "st %d", &sTime), sTime = 1000*sTime - 30, inc = 0, sTime /= unit;
 	else if(!strcmp(command, "name"))   { if(namOpt) EPRINT((f, "# setoption name UCI_Opponent value none none %s %s", comp ? "computer" : "human", line+5)) }
-	else if(!strcmp(command, "computer")) comp = 1;
-	else if(!strcmp(command, "offer"))  drawOffer = 1;
-	else if(!strcmp(command, "result")) { if(sc == 's') EPRINT((f, "# gameover %s\n", line[8] == '/' ? "draw" : (line[7] == '0') == mySide ? "win" : "lose")) }
-	else if(!strcmp(command, "quit"))   { EPRINT((f, "# quit\n")) fflush(toE), exit(0); }
+	else if(!strcmp(command, "result")) {
+	    if(sc == 's') EPRINT((f, "# gameover %s\n", line[8] == '/' ? "draw" : (line[7] == '0') == mySide ? "win" : "lose"))
+	    computer = NONE;
+	}
+	else if(!strcmp(command, "quit"))   { EPRINT((f, "# quit\n")) fflush(toE), exit(atoi(line+4)); }
+	else printf("Error (unknown command): %s\n", command);
+
+	fflush(stdout);
+}
+
+void *
+Engine2GUI()
+{
+    if(fromF = fopen("DefectiveEngineOptions.ini", "r")) printf("# fake engine input\n");
+    while(1) {
+	if(searching > 1) HandleEngineOutput();  // this could leave us (or fall through) pondering
+	while(qStart == qEnd && searching) HandleEngineOutput(); // relay ponder output until command arrives
+	Sync(PAUSE); // possibly wait for command silently if engine is idle
+	DoCommand(); LaunchSearch();
     }
 }
 
@@ -634,9 +731,10 @@ main(int argc, char **argv)
 {
 	char *dir = NULL, *p, *q; int e;
 
+
 	if(argc == 2 && !strcmp(argv[1], "-v")) { printf("UCI2WB " VERSION " by H.G.Muller\n"); exit(0); }
 	if(argc > 1 && !strcmp(argv[1], "debug")) { debug = 1; argc--; argv++; }
-	if(argc > 1 && !strcmp(argv[1], "-var")) { variants = argv[2]; argc-=2; argv+=2; }
+	if(argc > 1 && !strcmp(argv[1], "-var")) { strcpy(varList+1, argv[2]); *varList = ','; argc-=2; argv+=2; }
 	if(argc > 1 && argv[1][0] == '-') { sc = argv[1][1]; argc--; argv++; }
 	if(argc < 2) { printf("usage is: U%cI2WB [debug] [-s] <engine.exe> [<engine directory>]\n", sc-32); exit(-1); }
 	if(argc > 2) dir = argv[2];
@@ -647,7 +745,7 @@ main(int argc, char **argv)
 	else if(sc == 'n') sc = 'c'; // UCI for normal Chess
 
 	// spawn engine proc
-	if(StartEngine(argv[1], dir) != NO_ERROR) { perror(argv[1]), exit(-1); }
+	if(StartEngine(binary = argv[1], dir) != NO_ERROR) { perror(argv[1]), exit(-1); }
 
 	Sync(INIT);
 
